import unittest

from lift_utils import base, config, datatypes, lexicon, utils

LIFT_VERSION = config.LIFT_VERSION_LATEST


class TestEntry(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Entry()
        obj.id = lexicon.RefId("unique-entry-and-sense-id")
        obj.lexical_unit = lexicon.Multitext()
        obj.sense_items = [lexicon.Sense()]
        obj.date_created = None
        obj.sense_items[0].date_created = None

        xml = """
        <entry id="unique-entry-and-sense-id">
            <lexical-unit/>
            <sense/>
        </entry>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestEtymology(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Etymology(etym_type=datatypes.Key("Etype"), source="Esource")
        obj.gloss_items = [
            base.Gloss(lang="", text=""),
            base.Gloss(lang="", text=""),
        ]
        obj.form = base.Form(lang="", text="")

        xml = """
        <etymology type="Etype" source="Esource">
            <gloss lang="">
                <text/>
            </gloss>
            <gloss lang="">
                <text/>
            </gloss>
            <form lang="">
                <text/>
            </form>
        </etymology>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestExample(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Example()
        obj.source = base.Key("example source")
        obj.translation_items = [lexicon.Translation()]

        xml = """
        <example source="example source">
            <translation/>
        </example>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestExtensible(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Etymology()
        obj.type = datatypes.Key("Etype")
        obj.source = "Esource"
        obj.gloss_items = [
            base.Gloss(lang="", text=""),
            base.Gloss(lang="", text=""),
        ]
        obj.form = base.Form(lang="", text="")
        obj.date_created = datatypes.DateTime("0000-00-00")
        obj.date_modified = datatypes.DateTime("0000-00-00")
        obj.field_items = [
            base.Field(name=""),
            base.Field(name=""),
        ]
        obj.trait_items = [
            base.Trait(name="", value=""),
            base.Trait(name="", value=""),
        ]
        obj.annotation_items = [
            base.Annotation(name="", value=""),
            base.Annotation(name="", value=""),
        ]

        xml = f"""
        <etymology type="Etype" source="Esource" dateCreated="{obj.date_created}" dateModified="{obj.date_modified}">
            <gloss lang="">
                <text/>
            </gloss>
            <gloss lang="">
                <text/>
            </gloss>
            <form lang="">
                <text/>
            </form>
            <field name=""/>
            <field name= ""/>
            <trait name="" value=""/>
            <trait name="" value=""/>
            <annotation name="" value=""/>
            <annotation name="" value=""/>
        </etymology>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestGrammaticalInfo(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.GrammaticalInfo(value="g-i-value")
        obj.trait_items = [
            base.Trait(name="", value=""),
        ]

        xml = """
        <grammatical-info value="g-i-value">
            <trait name="" value=""/>
        </grammatical-info>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestLexicon(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Lexicon(version="0.13")
        obj.producer = f"LIFTUtils {config.LIB_VERSION}"
        obj.entry_items = [
            lexicon.Entry(),
        ]
        obj.entry_items[0].date_created = None

        xml = f"""
        <lift version="0.13" producer="LIFTUtils {config.LIB_VERSION}">
            <entry/>
        </lift>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestNote(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Note()
        obj.type = datatypes.Key("note-type")
        obj.form_items = [
            base.Form(lang=datatypes.Lang("en"), text=base.Text("This is a note"))
        ]

        xml = """
        <note type="note-type">
            <form lang="en">
                <text>This is a note</text>
            </form>
        </note>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestPhonetic(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        href1 = "https://example.com"
        href2 = "file:///home/user/audio.wav"
        obj = lexicon.Phonetic()
        obj.media_items = [
            base.URLRef(href=href1),
            base.URLRef(href=href2),
        ]
        obj.form_items = [
            base.Span(text=""),
            base.Span(text=""),
        ]

        xml = f"""
        <pronunciation>
            <form/>
            <form/>
            <media href="{href1}"/>
            <media href="{href2}"/>
        </pronunciation>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRelation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Relation(
            rel_type=base.Key("rel-type"), ref=lexicon.RefId("ref-id")
        )
        obj.usage_items = [base.Multitext()]

        xml = """
        <relation type="rel-type" ref="ref-id">
            <usage/>
        </relation>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestReversal(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Reversal()
        obj.main = lexicon.Reversal()
        obj.grammatical_info = lexicon.GrammaticalInfo(value="g-i-value")

        xml = """
        <reversal>
            <main/>
            <grammatical-info value="g-i-value"/>
        </reversal>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestSense(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Sense()
        obj.grammatical_info = lexicon.GrammaticalInfo()
        obj.definition = lexicon.Multitext()
        obj.subsense_items = [lexicon.Sense()]
        obj.date_created = None
        obj.subsense_items[0].date_created = None

        xml = """
        <sense>
            <grammatical-info/>
            <definition/>
            <subsense/>
        </sense>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestTranslation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Translation()
        obj.type = base.Key("literal")
        obj.form_items = [base.Form(lang="", text="")]

        xml = """
        <translation type="literal">
            <form lang="">
                <text/>
            </form>
        </translation>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestVariant(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = lexicon.Variant()
        obj.pronunciation_items = [lexicon.Phonetic()]
        obj.relation_items = [lexicon.Relation()]

        xml = """
        <variant>
            <pronunciation/>
            <relation/>
        </variant>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None



"""Define the basic datatypes."""

import uuid

from .utils import get_current_timestamp


class PCData(str):
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class DateTime(str):
    # format (str): YYYY-MM-DDTHH:MM:SSZZZZZZ
    # ZZZZZZ: +/-, H, H, :, M, M (offset from GMT)
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)
        else:
            return get_current_timestamp()


class Key(str):
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class Lang(str):
    # format (str): ISO[-SCRIPT[-x-PRIVATE]]
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)


class RefId(str):
    # format (HEX GUID): xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)
        else:  # generate new UUID
            return super().__new__(cls, str(uuid.uuid4()))


class URL(str):
    def __new__(cls, text=None):
        if text is not None:
            return super().__new__(cls, text)



import unittest

from lxml import etree

from lift_utils import config, lexicon

from . import DATA_PATH
from .utils import test_class_properties

LIFT_GOOD = str(DATA_PATH / "lexicon_good_v0.13_FW.lift")
LIFT_VERSION = "0.13"


class TestBaseNode(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//entry")
        self.entry = lexicon.Entry(xml_tree=self.xml_tree)

    def test_parent(self):
        self.assertIsNone(self.entry.parent_item)
        self.assertEqual(self.entry.etymology_items[0].parent_item, self.entry)


class TestEntry(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//entry")
        self.obj = lexicon.Entry(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestEtymology(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//etymology")
        self.obj = lexicon.Etymology(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestExample(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//example")
        self.obj = lexicon.Example(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestGrammaticalInfo(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//grammatical-info")  # noqa: E501
        self.obj = lexicon.GrammaticalInfo(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestLexicon(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot()
        self.obj = lexicon.Lexicon(path=LIFT_GOOD, xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def test_entry_items(self):
        self.assertTrue(len(self.obj.entry_items) > 0)

    def test_update_header_from_file(self):
        self.assertTrue(len(self.obj.header.ranges.range_items) > 1)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestNote(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//note")
        self.obj = lexicon.Note(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestPhonetic(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//pronunciation")  # noqa: E501
        self.obj = lexicon.Phonetic(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRelation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//relation")
        self.obj = lexicon.Relation(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestReversal(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//reversal")
        self.obj = lexicon.Reversal(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestSense(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//sense")
        self.obj = lexicon.Sense(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestTranslation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//translation")
        self.obj = lexicon.Translation(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestVariant(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(LIFT_GOOD).getroot().find(".//variant")
        self.obj = lexicon.Variant(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None



# flake8: noqa

import unittest
from lxml import etree

from .utils import compare_xml_trees

from lift_utils import base
from lift_utils import config
from lift_utils import datatypes
from lift_utils import header
from lift_utils import utils

LIFT_VERSION = "0.13"


class TestFieldDefn(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = header.FieldDefn(tag=base.Key("field-tag"))
        obj.form_items = [base.Form(lang="", text="")]

        xml = f"""
        <field tag="field-tag">
            <form lang="">
                <text/>
            </form>
        </field>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestFieldDefns(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = header.FieldDefns()
        obj.field_items = [
            header.FieldDefn(),
        ]

        xml = f"""
        <fields>
            <field/>
        </fields>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestHeader(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = header.Header()
        obj.ranges = header.Ranges()
        obj.ranges.range_items = [
            header.Range(),
            header.Range(),
        ]
        for r in obj.ranges.range_items:
            r.href = base.URL("///fake/path")
        obj.ranges.range_items[0].id = base.Key("range-1")
        obj.ranges.range_items[1].id = base.Key("range-2")
        obj.fields = header.FieldDefns()
        obj.fields.field_items = [
            header.FieldDefn(),
            header.FieldDefn(),
        ]
        obj.xml_tree = obj._to_xml_tree()

        xml = f"""
        <header>
            <ranges>
                <range id="range-1" href="///fake/path"/>
                <range id="range-2" href="///fake/path"/>
            </ranges>
            <fields>
                <field/>
                <field/>
            </fields>
        </header>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj.xml_tree
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRange(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = header.Range()
        obj.id = base.Key("range-id")
        obj.href = base.URL("///fake/path")
        obj.range_element_items = [header.RangeElement()]
        obj.xml_tree = obj._to_xml_tree()

        xml = f"""
        <range id="range-id" href="///fake/path">
            <range-element/>
        </range>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRangeElement(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = header.RangeElement(elem_id="range-element-id")
        obj.label_items = [base.Multitext()]

        xml = f"""
        <range-element id="range-element-id">
            <label/>
        </range-element>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRanges(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_lift_xml(self):
        obj = header.Ranges()
        obj.range_items = [
            header.Range(),
        ]
        obj.xml_tree = obj._to_xml_tree()
        xml = f"""
        <ranges>
            <range/>
        </ranges>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None



"""Manipulate lexicon entries and their dependent elements."""

from pathlib import Path
from typing import List, Optional, Union
from urllib.parse import unquote, urlparse

from lxml import etree

from . import config, utils
from .base import (
    Extensible,
    Form,
    Gloss,
    LIFTUtilsBase,
    Multitext,
    Trait,
    URLRef,
)
from .datatypes import URL, DateTime, Key, RefId
from .errors import InvalidExtensionError
from .header import Header, Range, Range13
from .utils import (
    get_writing_systems_from_entry,
    search_entry,
    xmlfile_to_etree,
)


class Note(Multitext, Extensible):
    """For storing descriptive information of many kinds.
    It can include comments, bibliographic information and domain specific
    notes.

    :ivar Optional[Key] type: Gives the type of note by reference to a
        ``range-element`` in the ``note-type`` range.
    """

    XML_TAG = "note"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        # TODO: Can we just use super().__init__() here and elsewhere?
        Extensible.__init__(self, **kwargs)
        Multitext.__init__(self, **kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(
            (
                "dateCreated",
                "dateModified",
                "type",
            )
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "field",
                "form",
                "pcdata",
                "span",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "type": Key,
            }
        )

        # attributes
        self.type: Optional[Key] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        return super().__str__()


class Phonetic(Multitext, Extensible):
    """This represents a single pronunciation in phonetic form.

    :ivar Optional[List[URLRef]] media_items: Stores an audio representation of
        the text.
    :ivar Optional[List[Span]] form_items: `Used by LIFT v0.13 (FieldWorks).`
        Stores the phonetic representation using whichever writing system: IPA,
        Americanist, etc.
    """

    XML_TAG = "pronunciation"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        Extensible.__init__(self, **kwargs)
        Multitext.__init__(self, **kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(
            (
                "dateCreated",
                "dateModified",
            )
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "field",
                "form",
                "media",
                "pcdata",
                "span",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "media": URLRef,
            }
        )

        # elements
        self.media_items: Optional[List[URLRef]] = None
        # NOTE: See note in base.Field definition re: superseded 'form_items'.
        # if config.LIFT_VERSION == '0.13':
        #     self.form_items: Optional[List[Span]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        return str(super())

    # NOTE: See note in base.Field definition re: superseded 'form_items'.
    # def add_form(
    #     self,
    #     text=None,
    #     tail=None,
    #     lang=None,
    #     href=None,
    #     span_class=None,
    # ):
    #     if config.LIFT_VERSION != '0.13':
    #         raise UnsupportedActionException(config.LIFT_VERSION)
    #     self._add_list_item(
    #         'form_items',
    #         Span,
    #         text=text,
    #         tail=tail,
    #         lang=lang,
    #         href=href,
    #         span_class=span_class,
    #     )

    # NOTE: See note in base.Field definition re: superseded 'form_items'.
    def add_form(self, lang, text):
        self._add_list_item(
            "form_items",
            Multitext,
            {lang: text},
        )

    def add_media(self, href=None, label=None):
        self._add_list_item("media_items", URLRef, href=href, label=label)


class Etymology(Extensible):
    """For describing lexical relations with a word not in the lexicon.

    :ivar Key type: Gives the etymological relationship between this sense and
        some other word in another language.
    :ivar str source: Gives the language for the source language of the
        etymological relation.
    :ivar Optional[List[Gloss]] gloss_items: Gives glosses of the word that the
        etymological relationship is with.
    :ivar Optional[Form] form: Holds the form of the etymological reference.
    """

    XML_TAG = "etymology"

    def __init__(
        self,
        etym_type: Key = None,
        source: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("source", "type"))
        self._attributes_optional = set(("dateCreated", "dateModified"))
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "field",
                "form",
                "gloss",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "type": Key,
                "source": str,
                "form": Form,
                "gloss": Gloss,
            }
        )

        # attributes
        self.type = etym_type
        self.source = source
        # elements
        self.gloss_items: Optional[List[Gloss]] = None
        self.form: Optional[Form] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        return f"{self.type} ({self.source})"

    def add_gloss(self, lang, text):
        self._add_list_item("gloss_items", Gloss, lang=lang, text=text)

    def set_form(self):
        pass


class GrammaticalInfo(LIFTUtilsBase):
    """A reference to a ``range-element`` in the ``grammatical-info`` range.

    :ivar Key value: The part of speech tag into the ``grammatical-info``
        range.
    :ivar Optional[List[Trait]] trait_items: Allows the grammatical information
        for a given sense to have more information than just the part of speech
        given by the ``value`` attribute.
    """

    XML_TAG = "grammatical-info"

    def __init__(
        self, value: Key = None, xml_tree: Optional[etree._Element] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("value",))
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set(("trait",))
        config.TAG_CLASSES.update(
            {
                "value": Key,
                "trait": Trait,
            }
        )

        # attributes
        self.value = None
        # elements
        self.trait_items: Optional[List[Trait]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        elif value is not None:
            self.value = Key(value)

    def __str__(self):
        traits = ""
        if self.trait_items:
            traits = f": {'; '.join([str(t) for t in self.trait_items])}"
        return f"{self.value}{traits}"


class Reversal(Multitext):
    """Enables a wider use of a dictionary.

    :ivar Optional[Key] type: Gives the type of the reversal as a
        ``range-element`` in the ``reversal-type`` range.
    :ivar Optional[Reversal] main: This gives the parent reversal in a
        hierarchical set of reversals.
    :ivar Optional[GrammaticalInfo] grammatical_info: This allows a reversal
        relation to specify what the grammatical information is in the reversal
        language.
    """

    XML_TAG = "reversal"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(("type",))
        self._elements_required = set()
        self._elements_optional = set(
            (
                "form",
                "grammatical-info",
                "main",
                "pcdata",
                "span",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "type": Key,
                "grammatical-info": GrammaticalInfo,
                "main": Reversal,
            }
        )

        # attributes
        self.type: Optional[Key] = None
        # elements
        self.main: Optional[Reversal] = None
        self.grammatical_info: Optional[GrammaticalInfo] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Translation(Multitext):
    """A ``Multitext`` with an optional translation ``type`` attribute.

    :ivar Optional[Key] type: Gives the type of the translation.
    """

    XML_TAG = "translation"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(("type",))
        self._elements_required = set()
        self._elements_optional = set(
            (
                "form",
                "pcdata",
                "span",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "type": Key,
            }
        )

        # attributes
        self.type: Optional[Key] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Example(Multitext, Extensible):
    """Gives an example sentence or phrase.
    It is given in the language and glosses of that example in other languages.

    :ivar Optional[Key] source: Reference by which another application may
        refer to this example or is a reference into another database of texts,
        for example.
    :ivar Optional[List[Translation]] translation_items: Gives translations of
        the example into different languages.
    :ivar Optional[List[Note]] note_items: Holds notes on this example.
    """

    XML_TAG = "example"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        # TODO: Consider replacing these two __inits__ with super().__init__.
        Extensible.__init__(self, **kwargs)
        Multitext.__init__(self, **kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(
            (
                "dateCreated",
                "dateModified",
                "source",
            )
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "field",
                "form",
                "note",
                "pcdata",
                "span",
                "trait",
                "translation",
            )
        )
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            self._elements_optional.discard("note")
        config.TAG_CLASSES.update(
            {
                "source": Key,
                "translation": Translation,
                "note": Note,
            }
        )

        # attributes
        self.source: Optional[Key] = None
        # elements
        self.translation_items: Optional[List[Translation]] = None
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            if hasattr(self, "note_items"):
                del self.note_items
        else:
            self.note_items: Optional[List[Note]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Relation(Extensible):
    """This element is used for lexical relations.

    :ivar Key type: Is the type of the particular lexical relation.
    :ivar RefId ref: This is the other end of the relation, either a ``sense``
        or an ``entry``.
    :ivar Optional[int] order: Gives the relative ordering of relations of a
        given type when a multiple relation is being described.
    :ivar Optional[Multitext] usage: Gives information on usage in one or more
        languages or writing systems.
    """

    XML_TAG = "relation"

    def __init__(
        self,
        rel_type: Key = None,
        ref: RefId = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("ref", "type"))
        self._attributes_optional = set(
            (
                "dateCreated",
                "dateModified",
                "order",
            )
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "field",
                "trait",
                "usage",
            )
        )
        config.TAG_CLASSES.update(
            {
                "type": Key,
                "ref": RefId,
                "order": int,
                "usage": Multitext,
            }
        )

        # attributes
        self.type = rel_type
        self.ref = ref
        self.order: Optional[int] = None
        # elements
        self.usage_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        return f"{self.type}: {self.ref}"


class Variant(Multitext, Extensible):
    """``Variant`` elements are used for all sorts of variation.

    :ivar Optional[RefId] ref: Gives the variation as a reference to another
        ``entry`` or ``sense`` rather than specifying the ``form`` (that is,
        the ``Multitext`` value of the variant).
    :ivar Optional[List[Phonetic]] pronunciation_items: Holds the phonetic
        variant whether it is that this is a variation in phonetics only or
        that the phonetic variation arises because of an orthographic or
        phonemic variation.
    :ivar Optional[List[Relation]] relation_items: Some variants have a lexical
        relationship with other senses or entries in the lexicon.
    """

    XML_TAG = "variant"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        Extensible.__init__(self, **kwargs)
        Multitext.__init__(self, **kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(
            (
                "dateCreated",
                "dateModified",
                "ref",
            )
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "field",
                "form",
                "pcdata",
                "pronunciation",
                "relation",
                "span",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "ref": RefId,
                "pronunciation": Phonetic,
                "relation": Relation,
            }
        )

        # attributes
        self.ref: Optional[RefId] = None
        # elements
        self.pronunciation_items: Optional[List[Phonetic]] = None
        self.relation_items: Optional[List[Relation]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        return self.ref if self.ref else "variant"


class Sense(Extensible):
    """An ``entry`` is made up of a number of ``sense`` elements.

    :ivar Optional[RefId] id: This gives an identifier for this ``Sense`` so
        that things can refer to it.
    :ivar Optional[int] order: A number that is used to give the relative
        order of senses within an entry.
    :ivar Optional[GrammaticalInfo] grammatical_info: Grammatical information.
    :ivar Optional[List[Union[Gloss, Form]]] gloss_items: Each ``gloss`` is a
        single string in a single language and writing system. ``Form`` is used
        by LIFT v0.13 (FieldWorks), while ``Gloss`` is used in later versions.
    :ivar Optional[Multitext] definition: Gives the definition in multiple
        languages or writing systems.
    :ivar Optional[List[Relation]] relation_items: While a lexical relation
        isn't strictly owned by a sense it is a good place to hold it.
    :ivar Optional[List[Note]] note_items: There are lots of different types of
        notes.
    :ivar Optional[List[Example]] example_items: Examples may be used for
        different target audiences.
    :ivar Optional[List[Reversal]] reversal_items: There may be different
        reversal indexes.
    :ivar Optional[List[URLRef]] illustration_items: The picture doesn't have
        to be static.
    :ivar Optional[List[Sense]] subsense_items: Senses can form a hierarchy.
    """

    XML_TAG = "sense"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(
            (
                "dateCreated",
                "dateModified",
                "id",
                "order",
            )
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "definition",
                "example",
                "field",
                "gloss",
                "grammatical-info",
                "illustration",
                "note",
                "relation",
                "reversal",
                "subsense",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "id": RefId,
                "order": int,
                "definition": Multitext,
                "example": Example,
                "gloss": Gloss,
                "grammatical-info": GrammaticalInfo,
                "illustration": URLRef,
                "note": Note,
                "relation": Relation,
                "reversal": Reversal,
                "subsense": Sense,
            }
        )

        # attributes
        self.id: Optional[RefId] = None
        self.order: Optional[int] = None
        # elements
        self.definition: Optional[Multitext] = None
        self.example_items: Optional[List[Example]] = None
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            config.TAG_CLASSES["gloss"] = Form
            self.gloss_items: Optional[List[Form]] = None
        else:
            self.gloss_items: Optional[List[Gloss]] = None
        self.grammatical_info: Optional[GrammaticalInfo] = None
        self.illustration_items: Optional[List[URLRef]] = None
        self.note_items: Optional[List[Note]] = None
        self.relation_items: Optional[List[Relation]] = None
        self.reversal_items: Optional[List[Reversal]] = None
        self.subsense_items: Optional[List[Sense]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        else:
            self.set_date_created()

    def __str__(self):
        return self._summary_line()

    def add_example(self) -> Example:
        """Add an empty ``Example`` item to the sense.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("example_items", Example)

    def add_gloss(self, lang, text) -> Gloss:
        """Add a ``Gloss`` item to the sense.
        Returns the new object.

        :var str lang: The gloss's language.
        :var str text: The actual gloss text.
        """
        self.set_date_modified()
        return self._add_list_item("gloss_items", Gloss, lang=lang, text=text)

    def add_illustration(self, href=None) -> URLRef:
        """Add a ``URLRef`` illustration item to the sense.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("illustration_items", URLRef, href=href)

    def add_note(self) -> Note:
        """Add an empty ``Note`` item to the sense.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("note_items", Note)

    def add_relation(self) -> Relation:
        """Add an empty ``Relation`` item to the sense.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("relation_items", Relation)

    def add_reversal(self) -> Reversal:
        """Add an empty ``Reversal`` item to the sense.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("reversal_items", Reversal)

    def add_subsense(self):
        """Add an empty ``Subense`` item to the sense.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("subsense_items", Sense)

    def get_id(self) -> RefId:
        """Return the object's ``id`` attribute."""
        return self.id

    def get_gloss(self, lang="en") -> str:
        """Get the gloss for a given language.
        Defaults to English; falls back otherwise to the first gloss.
        """
        gloss = ""
        if self.gloss_items:
            # Choose 1st gloss if preferred language not found.
            gloss = str(self.gloss_items[0])
            for g in self.gloss_items:
                if g.lang == lang:
                    # Choose preferred language gloss.
                    gloss = str(g)
        return gloss

    def get_grammatical_info(self) -> GrammaticalInfo:
        """Return the grammatical-info part of speech value."""
        grammatical_info = ""
        if self.grammatical_info:
            grammatical_info = str(self.grammatical_info)
        return grammatical_info

    def set_definition(self, forms_dict=None):
        """Set the sense's definition.

        :var Optional[dict] forms_dict: ``dict`` keys are language codes,
            values are the text for each definition.
        """
        self.set_date_modified()
        self.definition = Multitext(forms_dict)

    def set_grammatical_info(self, value: str):
        """Set the sense's ``GrammaticalInfo``.

        :var str value: The part of speech tag in the ``grammatical-info``
            range.
        """
        self.set_date_modified()
        self.grammatical_info = GrammaticalInfo(value=value)

    def _summary_line(self, lang="en"):
        """Return a one-line summary of the entry's data for a given language.
        Defaults to English.
        """
        gl = utils.ellipsize(str(self.get_gloss(lang=lang)), 20)
        gi = utils.ellipsize(self.get_grammatical_info(), 10)
        return f"{gl:20}\t{gi:10}\t{self.id}"


class Entry(Extensible):
    """This is the core of a lexicon. A lexicon is made up of a set of entries.

    :ivar Optional[RefId] id: This gives a unique identifier to this ``entry``.
    :ivar Optional[str] guid: `Deprecated.` This gives a unique identifier to
        this entry in the form of a “universally unique identifier” (RFC 4122).
    :ivar Optional[int] order: This is the homograph number.
    :ivar Optional[DateTime] date_deleted: If this attribute exists then it
        indicates that the particular ``entry`` has been deleted.
    :ivar Optional[Multitext] lexical_unit: The lexical form is the primary
        lexical form as is found as the primary lexical form in the source data
        models for this standard.
    :ivar Optional[Multitext] citation: This is the form that is to be printed
        in the dictionary.
    :ivar Optional[List[Phonetic]] pronunciation_items: There can be multiple
        phonetic forms of an entry.
    :ivar Optional[List[Variant]] variant_items: Any constrained variants or
        free orthographic variants.
    :ivar Optional[List[Sense]] sense_items: This is where the definition goes.
    :ivar Optional[List[Note]] note_items: The more notes you keep the better.
    :ivar Optional[List[Relation]] relation_items: Gives a lexical relationship
        between this entry and another ``entry`` or ``sense``.
    :ivar Optional[List[Etymology]] etymology_items: Differs from a lexical
        relation in that it has no referent in the lexicon.
    """

    XML_TAG = "entry"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(
            ("dateCreated", "dateDeleted", "dateModified", "guid", "id", "order")
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "citation",
                "etymology",
                "field",
                "lexical-unit",
                "note",
                "pronunciation",
                "relation",
                "sense",
                "trait",
                "variant",
            )
        )
        config.TAG_CLASSES.update(
            {
                "guid": str,
                "id": RefId,
                "dateDeleted": DateTime,
                "order": int,
                "citation": Multitext,
                "etymology": Etymology,
                "lexical-unit": Multitext,
                "note": Note,
                "pronunciation": Phonetic,
                "relation": Relation,
                "sense": Sense,
                "variant": Variant,
            }
        )

        # attributes
        self.id: Optional[RefId] = None
        self.guid: Optional[str] = None  # deprecated
        self.date_deleted: Optional[DateTime] = None
        self.order: Optional[int] = None
        # elements
        self.citation: Optional[Multitext] = None
        self.etymology_items: Optional[List[Etymology]] = None
        self.lexical_unit: Optional[Multitext] = None
        self.note_items: Optional[List[Note]] = None
        self.pronunciation_items: Optional[List[Phonetic]] = None
        self.relation_items: Optional[List[Relation]] = None
        self.sense_items: Optional[List[Sense]] = None
        self.variant_items: Optional[List[Variant]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        else:
            self.set_date_created()

    def __str__(self):
        return self._summary_line()

    def add_etymology(self) -> Etymology:
        """Add an empty ``Etymology`` item to the entry.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("etymology_items", Etymology)

    def add_note(self) -> Note:
        """Add an empty ``Note`` item to the entry.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("note_items", Note)

    def add_pronunciation(self) -> Phonetic:
        """Add an empty ``Phonetic`` item to the entry.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("pronunciation_items", Phonetic)

    def add_relation(self) -> Relation:
        """Add an empty ``Relation`` item to the entry.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("relation_items", Relation)

    def add_sense(self) -> Sense:
        """Add an empty ``Sense`` item to the entry.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        sense = self._add_list_item("sense_items", Sense)
        sense.date_created = DateTime()
        new_id = RefId()
        # NOTE: When creating an entry from a lexicon we can check all existing
        # IDs to make sure the generated one is unique. But when creating a
        # sense from an entry we don't have the ability to access the lexicon
        # directly. Will just assume for now that the generated ID is unique.
        sense.id = new_id
        return sense

    def add_variant(self) -> Variant:
        """Add an empty ``Variant`` item to the entry.
        Returns the new object, which can then be used to add data to it.
        """
        self.set_date_modified()
        return self._add_list_item("variant_items", Variant)

    def get_id(self) -> RefId:
        """Return the object's unique identifier"""
        return self.id

    def set_citation(self, forms_dict=None):
        """Set the entry's ``Citation``.

        :var Optional[dict] forms_dict: ``dict`` keys are language codes,
            values are the text descriptions of the ``Citation``.
        """
        self.set_date_modified()
        self.citation = Multitext(forms_dict)

    def set_lexical_unit(self, forms_dict=None):
        """Set the entry's ``LexicalUnit``.

        :var Optional[dict] forms_dict: ``dict`` keys are lanuage codes,
            values are text descriptions of the ``LexicalUnit``.
        """
        self.set_date_modified()
        self.lexical_unit = Multitext(forms_dict)

    def _summary_line(self, lang="en"):
        """Return a one-line summary of the entry's data for a given language.
        Defaults to English.
        """
        lu = utils.ellipsize(str(self.lexical_unit), 20)
        # gl = self.get_gloss(lang=lang)
        # return f"{lu:20}\t{gl:30}\t{self.id}"
        return f"{lu:20}\t{self.id}"


class Lexicon(LIFTUtilsBase):
    """This is the main class of the lexicon.
    It contains the header and all the entries in the database.

    :ivar str version: Specifies the lift language version number.
    :ivar Optional[str] producer: Identifies the particular producer of this
        lift file.
    :ivar Optional[Header] header: Contains the header information for the
        database.
    :ivar Optional[List[Entry]] entry_items: Each of the entries in the
        lexicon.
    :ivar Optional[Path] path: File path to a LIFT file to import.
    """

    XML_TAG = "lift"

    def __init__(
        self,
        path: Optional[Union[Path, str]] = None,
        version: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("version",))
        self._attributes_optional = set(("producer",))
        self._elements_required = set()
        self._elements_optional = set(("entry", "header"))
        config.TAG_CLASSES.update(
            {
                "version": str,
                "producer": str,
                "entry": Entry,
                "header": Header,
            }
        )

        self.lift_xml_tree = None
        self.ranges_xml_tree = None
        self.analysis_writing_systems = None
        self.vernacular_writing_systems = None
        # attributes
        self.version = version
        # Make version accessible globally.
        config.LIFT_VERSION = self.version
        self.producer: str = f"LIFT-Utils {config.LIB_VERSION}"
        # elements
        self.header: Optional[Header] = None
        self.entry_items: Optional[List[Entry]] = None
        if path:
            self.path = Path(path).expanduser()
            if self.path.suffix == ".lift":
                self._from_lift(self.path)
            else:
                raise InvalidExtensionError(self.path.name)
        elif xml_tree is not None:
            self._from_xml_tree(xml_tree)
            self._find_writing_systems()

    def __str__(self):
        return f"LIFT lexicon v{self.version}; produced by {self.producer}"

    def add_entry(self) -> Entry:
        """Add an empty ``Entry`` to the lexicon.
        Returns the ``Entry`` object, which can then be used to add data to it.
        """
        entry = self._add_list_item("entry_items", Entry)
        entry.date_created = DateTime()
        new_id = RefId()
        while self.get_item_by_id(new_id):  # make sure it's unique
            new_id = RefId()
        entry.id = new_id
        return entry

    def find(
        self, text: str, field: str = "gloss", match_type: str = "contains"
    ) -> Union[Entry, Sense, None]:
        """Return the first matching ``Entry`` or ``Sense`` item.
        The field searched can be the entry's "lexical-unit" or "variant"
        field, the sense's "gloss" [default], "definition", or
        "grammatical-info" field, as well as any fields defined in the LIFT
        file's header.

        :var str text: The search term.
        :var str field: The field to be searched [default is "gloss"].
        :var str match_type: The kind comparison between the search term and
            the field's data. Possible values are "contains" [default],
            "exact", or "regex".
        """
        return self._find(text, field=field, match_type=match_type)

    def find_all(
        self, text: str = "", field: str = "gloss", match_type: str = "contains"
    ) -> List[Union[Entry, Sense]]:
        """Return all matching ``Entry`` or ``Sense`` items.
        The field searched can be the entry's "lexical-unit" or "variant"
        field, the sense's "gloss" [default], "definition", or
        "grammatical-info" field, as well as any fields defined in the LIFT
        file's header.

        :var str text: The search term.
        :var str field: The field to be searched [default is "gloss"].
        :var str match_type: The kind comparison between the search term and
            the field's data. Possible values are "contains" [default],
            "exact", or "regex".
        """
        return self._find(text, field=field, match_type=match_type, get_all=True)

    def get_item_by_id(self, refid: str) -> Union[Entry, Sense, None]:
        """Return an entry or sense by its ``id`` attribute.

        :var str refid: The ``id`` attribute of the entry or sense.
        """
        return self._item_from_id(refid, item_type="self")

    def get_range_elements(self, range_name):
        """Returns a generator object that lists all the names defined in the
        header for the given ``range``.

        :var str range_name: The name of the header range.
        """
        for r in self.header.ranges.range_items:
            if r.id == range_name:
                for e in r.range_element_items:
                    yield e.id

    def get_ranges(self):
        """Returns a generator object that lists all the range names defined in
        the header.
        """
        for r in self.header.ranges.range_items:
            yield r.id

    def show(self):
        """Print an overview of the ``Lexicon`` in the terminal window."""
        text = None
        if self.entry_items:
            summary_lines = [e._summary_line("en") for e in self.entry_items]  # noqa: E501
            slist = utils.unicode_sort(summary_lines)
            text = "\n".join(slist)
        print(text)

    def to_lift(self, file_path: str):
        """Save the ``Lexicon`` as a LIFT file.
        The LIFT-RANGES file will be automatically created in the same folder
        as the LIFT file.

        :var str file_path: Full or relative path to new LIFT file.
        """
        outfile = (
            Path(file_path).expanduser().with_suffix(".lift")
        )  # ensure suffix  # noqa: E501
        ranges_file = outfile.with_suffix(".lift-ranges")

        # Write LIFT file.
        lift_tree = self._to_xml_tree()
        for _range in lift_tree.find(".//ranges").getchildren():
            for _range_element in _range:
                _range.remove(_range_element)
            for attrib in _range.attrib.keys():
                if attrib not in ["id", "href"]:
                    del _range.attrib[attrib]
        outfile.write_text(self._to_xml(lift_tree))

        # Write LIFT-RANGES file.
        lift_ranges = self.header.ranges._to_xml_tree()
        lift_ranges.tag = "lift-ranges"
        for _range in list(lift_ranges):
            del _range.attrib["href"]
        ranges_file.write_text(self._to_xml(lift_ranges))

    def _find(self, text, field="gloss", match_type="contains", get_all=False):
        items = []
        target_groups = ["entries", "senses"]
        entry_only_fields = ["lexical-unit", "variant"]
        sense_only_fields = ["gloss", "definition", "grammatical-info"]
        header_fields = [f.tag for f in self.header.fields.field_items]
        if field in entry_only_fields:
            target_groups.remove("senses")
        elif field in sense_only_fields:
            target_groups.remove("entries")

        for entry in self.entry_items:
            result = search_entry(
                entry, text, field, target_groups, header_fields, match_type, get_all
            )
            if result is not None:
                if not get_all:
                    return result
                else:
                    items.extend(result)
        if get_all:
            return items

    def _find_writing_systems(self):
        """Infer the lexicon's writing systems (vernacular and analysis)."""
        if self.vernacular_writing_systems is None:
            self.vernacular_writing_systems = []
        if self.analysis_writing_systems is None:
            self.analysis_writing_systems = []

        for e in self.entry_items:
            writing_systems = get_writing_systems_from_entry(e)
            for lang in writing_systems.get("vernacular"):
                if lang not in self.vernacular_writing_systems:
                    self.vernacular_writing_systems.append(lang)
            for lang in writing_systems.get("analysis"):
                if lang not in self.analysis_writing_systems:
                    self.analysis_writing_systems.append(lang)

    def _from_lift(self, infile):
        infile = Path(infile)
        if not infile.is_file():
            raise FileNotFoundError
        self._from_xml_tree(xmlfile_to_etree(infile))
        self._find_writing_systems()

    def _item_from_id(self, refid, item_type="self"):
        if not self.entry_items:
            return
        for entry in self.entry_items:
            if entry.id == refid:
                if item_type == "self":
                    return entry
                elif item_type == "parent":
                    return None
            if entry.sense_items:
                for sense in entry.sense_items:
                    if sense.id == refid:
                        if item_type == "self":
                            return sense
                        elif item_type == "parent":
                            return entry
                    if sense.subsense_items:
                        for subsense in sense.subsense_items:
                            if subsense.id == refid:
                                if item_type == "self":
                                    return subsense
                                elif item_type == "parent":
                                    return sense

    def _from_xml_tree(self, xml_tree):
        self.version = xml_tree.attrib.get("version")
        # Allow global access to version number.
        config.LIFT_VERSION = self.version
        # Update object attributes.
        super()._from_xml_tree(xml_tree)
        # Update header range data from external file(s).
        ext_hrefs = set()
        for r in self.header.ranges.range_items:
            if r.href:
                ext_hrefs.add(r.href)
        for p in ext_hrefs:
            self._update_header_from_href(p)
        # Update ranges xml_trees.
        self.header.ranges._to_xml_tree()

    def _update_header_from_href(self, href: URL):
        filepath = unquote(urlparse(href).path)
        try:
            xml_tree = xmlfile_to_etree(filepath)
        except OSError:
            # Probably absolute URI from a different device.
            # Try same file name, but in same dir as current LIFT file.
            relpath = self.path.parent / Path(filepath).name
            xml_tree = xmlfile_to_etree(relpath)

        # Update model data.
        for _range in xml_tree.getchildren():
            for i, r in enumerate(self.header.ranges.range_items[:]):
                if _range.attrib.get("id") == r.id:
                    if self.version == config.LIFT_VERSION_FIELDWORKS:
                        self.header.ranges.range_items[i] = Range13(xml_tree=_range)  # noqa: E501
                    else:
                        self.header.ranges.range_items[i] = Range(xml_tree=_range)  # noqa: E501
                    self.header.ranges.range_items[i].href = r.href  # add href
                    break



import unittest

from lxml import etree

from lift_utils import base, config

from . import DATA_PATH
from .utils import test_class_properties, test_properties

ENTRY_LIFT_GOOD = str(DATA_PATH / "entry_good_v0.15.lift")
LIFT_VERSION = config.LIFT_VERSION_LATEST


class TestAnnotation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//annotation")  # noqa: E501
        self.obj = base.Annotation(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestExtensible(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.obj = base.Extensible(etree.parse(ENTRY_LIFT_GOOD).getroot())

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestField(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//field")
        self.obj = base.Field(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestForm(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//form")
        self.obj = base.Form(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestGloss(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//gloss")
        self.obj = base.Gloss(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestMultitext(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//annotation")  # noqa: E501
        self.obj = base.Multitext(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestSpan(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//span")
        self.obj = base.Span(xml_tree=self.xml_tree)

    def test_properties(self):
        for group in (
            self.obj._attributes_required,
            self.obj._elements_required,
        ):
            test_properties(self, group, optional=False)
        elements_optional = self.obj._elements_optional.copy()
        elements_optional.discard("span")  # hard to test optional nested element
        for group in (
            self.obj._attributes_optional,
            elements_optional,
        ):
            test_properties(self, group, optional=True)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestText(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//form")
        self.obj = None
        for c in self.xml_tree.getchildren():
            if c.tag == "text":
                self.obj = base.Text(c)
                break
        if not self.obj:
            raise AssertionError

    def test_properties(self):
        for group in (
            self.obj._attributes_required,
            self.obj._elements_required,
        ):
            test_properties(self, group, optional=False)
        elements_optional = self.obj._elements_optional.copy()
        elements_optional.discard("span")  # already tested in TestSpan
        for group in (
            self.obj._attributes_optional,
            elements_optional,
        ):
            test_properties(self, group, optional=True)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestTrait(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find("trait")
        self.obj = base.Trait(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestURLRef(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(ENTRY_LIFT_GOOD).getroot().find(".//urlref")  # noqa: E501
        self.obj = base.URLRef(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None



import unittest

from lxml import etree

from lift_utils import config, header

from . import DATA_PATH
from .utils import test_class_properties, test_properties

HEADER_LIFT_GOOD = str(DATA_PATH / "header_good_v0.15.lift")
LIFT_VERSION = config.LIFT_VERSION_LATEST


class TestFieldDefinition(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find(".//field")  # noqa: E501
        self.obj = header.FieldDefinition(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestFields(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find(".//fields")  # noqa: E501
        self.obj = header.Fields(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def test_items(self):
        self.assertTrue(len(self.obj.field_items) > 1)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestHeader(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find(".//header")  # noqa: E501
        self.obj = header.Header(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRanges(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find(".//ranges")  # noqa: E501
        self.obj = header.Ranges(xml_tree=self.xml_tree)

    def test_properties(self):
        test_class_properties(self)

    def test_items(self):
        self.assertTrue(len(self.obj.range_items) > 1)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRange(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find(".//range")  # noqa: E501
        self.obj = header.Range(xml_tree=self.xml_tree)

    def test_properties(self):
        for group in (
            self.obj._attributes_required,
            self.obj._elements_required,
        ):
            test_properties(self, group, optional=False)
        elem_optional = self.obj._elements_optional.copy()
        elem_optional.discard("annotation")
        elem_optional.discard("field")
        elem_optional.discard("trait")
        for group in (
            self.obj._attributes_optional,
            elem_optional,
        ):
            test_properties(self, group, optional=True)

    def tearDown(self):
        config.LIFT_VERSION = None


class TestRangeElement(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION
        self.xml_tree = etree.parse(HEADER_LIFT_GOOD).getroot().find(".//range-element")  # noqa: E501
        self.obj = header.RangeElement(xml_tree=self.xml_tree)

    def test_properties(self):
        for group in (
            self.obj._attributes_required,
            self.obj._elements_required,
        ):
            test_properties(self, group, optional=False)
        elem_optional = self.obj._elements_optional.copy()
        elem_optional.discard("annotation")
        elem_optional.discard("field")
        elem_optional.discard("trait")
        for group in (
            self.obj._attributes_optional,
            elem_optional,
        ):
            test_properties(self, group, optional=True)

    def tearDown(self):
        config.LIFT_VERSION = None



"""Define runtime global variables."""

from lxml import etree

LIB_VERSION = "0.4"
LIFT_VERSION_FIELDWORKS = "0.13"  # used by FLEx 9.1
LIFT_VERSION_LATEST = "0.15"
LIFT_VERSION_DEFAULT = LIFT_VERSION_FIELDWORKS
LIFT_VERSION = None

TAG_CLASSES = dict()

XML_PARSER = etree.XMLParser(remove_blank_text=True)



"""Functions needed to run demo scripts."""

from .lexicon import Lexicon


def lexical_units_from_lift(lift, cawls):
    lex = Lexicon(lift)
    lexical_units = []
    for cawl in cawls:
        lexical_unit = None
        item = lex.find(cawl, field="CAWL", match_type="exact")
        if item and item.id:
            parent = item.parent_item
            if parent:
                lexical_unit = parent.lexical_unit
        lexical_units.append(lexical_unit)
    return lexical_units



"""Manipulate the header section."""

from typing import List, Optional

from lxml import etree

from . import config
from .base import Extensible, LIFTUtilsBase, Multitext
from .datatypes import URL, Key


class FieldDefn(Multitext):
    """Gives information about a particular field type.
    It may be used by an application to add information not part of the LIFT
    standard.

    .. note:: Used by LIFT v0.13 (FieldWorks) instead of ``FieldDefinition``.
    """

    XML_TAG = "field"

    def __init__(
        self, tag: Key = None, xml_tree: Optional[etree._Element] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("tag",))
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set(("form", "pcdata", "span", "trait"))
        config.TAG_CLASSES.update(
            {
                "tag": Key,
            }
        )

        # attributes
        self.tag = tag

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        forms = "forms" if len(self.form_items) > 1 else "form"
        return f"{self.tag} ({len(self.form_items)} {forms})"


class FieldDefinition(LIFTUtilsBase):
    """Gives information about a particular field type.
    It may be used by an application to add information not part of the LIFT
    standard.
    """

    XML_TAG = "field"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set(("name",))
        self._attributes_optional = set(
            ("class", "option-range", "type", "writing-system")
        )
        self._elements_required = set()
        self._elements_optional = set(("label", "description"))
        config.TAG_CLASSES.update(
            {
                "name": Key,
                "class": str,
                "option-range": Key,
                "type": str,
                "writing-system": str,
                "label": Multitext,
                "description": Multitext,
            }
        )

        # attributes
        self.name: Key = None
        self.class_: Optional[str] = None
        self.type: Optional[str] = None
        self.option_range: Optional[Key] = None
        self.writing_system: Optional[str] = None
        # elements
        self.label: Optional[Multitext] = None
        self.description: Optional[Multitext] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        return self.name


class FieldDefns(LIFTUtilsBase):
    """This is a simple list of ``field-defn`` elements.

    .. note:: Used by LIFT v0.13 (FieldWorks) instead of ``Fields``.
    """

    XML_TAG = "fields"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set(("field",))
        config.TAG_CLASSES.update(
            {
                "field": FieldDefn,
            }
        )

        # elements
        self.field_items: Optional[List[FieldDefn]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Fields(LIFTUtilsBase):
    """This is a simple list of ``field-definition`` elements."""

    # FIXME: Should this class have an XML tag?
    XML_TAG = None

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set(("field",))
        config.TAG_CLASSES.update(
            {
                "field": FieldDefinition,
            }
        )

        # elements
        self.field_items: Optional[List[FieldDefinition]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class RangeElement13(LIFTUtilsBase):
    """The description of a particular range element found in a ``range``."""

    XML_TAG = "range-element"

    def __init__(
        self, elem_id: Key = None, xml_tree: Optional[etree._Element] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("id",))
        self._attributes_optional = set(("guid", "parent"))
        self._elements_required = set()
        self._elements_optional = set(
            (
                "abbrev",
                "description",
                "label",
            )
        )
        config.TAG_CLASSES.update(
            {
                "id": Key,
                "guid": str,
                "parent": Key,
                "abbrev": Multitext,
                "description": Multitext,
                "label": Multitext,
            }
        )

        # attributes
        self.id = elem_id
        self.parent: Key = None
        self.guid: str = None
        # elements
        self.description_items: Optional[List[Multitext]] = None
        self.label_items: Optional[List[Multitext]] = None
        self.abbrev_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class RangeElement(Extensible):
    """The description of a particular range element found in a ``range``.

    .. note:: Does not inherit from Extensible in LIFT v0.13 (FieldWorks).
    """

    XML_TAG = "range-element"

    def __init__(
        self, elem_id: Key = None, xml_tree: Optional[etree._Element] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("id",))
        self._attributes_optional = set(
            ("dateCreated", "dateModified", "guid", "parent")
        )
        self._elements_required = set()
        self._elements_optional = set(
            (
                "annotation",
                "abbrev",
                "description",
                "field",
                "label",
                "trait",
            )
        )
        config.TAG_CLASSES.update(
            {
                "id": Key,
                "guid": str,
                "parent": Key,
                "abbrev": Multitext,
                "description": Multitext,
                "label": Multitext,
            }
        )

        # attributes
        self.id = elem_id
        self.parent: Key = None
        self.guid: str = None
        # elements
        self.description_items: Optional[List[Multitext]] = None
        self.label_items: Optional[List[Multitext]] = None
        self.abbrev_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Range13(LIFTUtilsBase):
    """A set of ``range-elements``.
    It is used to identify both the group of ``range-elements`` but also to
    some extent their type.
    """

    XML_TAG = "range"

    def __init__(
        self, range_id: Key = None, xml_tree: Optional[etree._Element] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("id",))
        self._attributes_optional = set(("guid", "href"))
        self._elements_required = set(("range-element",))
        self._elements_optional = set(
            (
                "abbrev",
                "description",
                "label",
            )
        )
        config.TAG_CLASSES.update(
            {
                "id": Key,
                "guid": str,
                "href": URL,
                "range-element": RangeElement13,
                "abbrev": Multitext,
                "description": Multitext,
                "label": Multitext,
            }
        )

        # attributes
        self.id = range_id
        self.guid: Optional[str] = None
        self.href: Optional[URL] = None
        # elements
        self.range_element_items: List[RangeElement13] = None
        self.abbrev_items: Optional[List[Multitext]] = None
        self.description: Optional[Multitext] = None
        self.label_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Range(Extensible):
    """A set of ``range-elements``.
    It is used to identify both the group of ``range-elements`` but also to
    some extent their type.
    """

    XML_TAG = "range"

    def __init__(
        self, range_id: Key = None, xml_tree: Optional[etree._Element] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("id",))
        self._attributes_optional = set(("dateCreated", "dateModified", "guid", "href"))
        self._elements_required = set(("range-element",))
        self._elements_optional = set(
            ("abbrev", "annotation", "description", "field", "label", "trait")
        )
        config.TAG_CLASSES.update(
            {
                "id": Key,
                "guid": str,
                "href": URL,
                "range-element": RangeElement,
                "abbrev": Multitext,
                "description": Multitext,
                "label": Multitext,
            }
        )

        # attributes
        self.id = range_id
        self.guid: Optional[str] = None
        self.href: Optional[URL] = None
        # elements
        self.range_element_items: List[RangeElement] = None
        self.abbrev_items: Optional[List[Multitext]] = None
        self.description: Optional[Multitext] = None
        self.label_items: Optional[List[Multitext]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Ranges(LIFTUtilsBase):
    """The root element in a Lift Ranges file."""

    XML_TAG = "ranges"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set(("range",))
        self._elements_optional = set()
        config.TAG_CLASSES.update(
            {
                "range": Range,
            }
        )

        # elements
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            config.TAG_CLASSES["range"] = Range13
            self.range_items: List[Range13] = None
        else:
            self.range_items: List[Range] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)


class Header(LIFTUtilsBase):
    """The header information for a LIFT file.
    It includes range information and added field definitions.

    :ivar Optional[Multitext] description: Contains a multilingual description
        of the lexicon for information purposes only.
    :ivar Optional[Ranges] ranges: Contains all the ``range`` information.
    :ivar Optional[FieldDefns] field_items: `Used by LIFT v0.13 (FieldWorks).`
        Contains definitions for all the ``field`` types used in the document.
    :ivar Optional[Fields] fields: Contains definitions for all the ``field``
        types used in the document.
    """

    XML_TAG = "header"

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set(("description", "fields", "ranges"))
        self._elements_optional = set()
        config.TAG_CLASSES.update(
            {
                "description": Multitext,
                "fields": Fields,
                "ranges": Ranges,
            }
        )

        # elements
        self.description: Optional[Multitext] = None
        self.ranges: Optional[Ranges] = None
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            config.TAG_CLASSES["fields"] = FieldDefns
            self.fields: Optional[FieldDefns] = None
        else:
            self.fields: Optional[Fields] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def __str__(self):
        s = "Header"
        if self.description:
            s = str(self.description)
        if self.ranges and self.ranges.range_items:
            s += f": {len(self.range_items)} ranges"
        if self.fields:
            s += f", {len(self.fields)} fields"
        return s



"""Various utility functions."""

import re
from datetime import datetime, timezone

import unidecode
from lxml import etree

from . import config


def ellipsize(string, length):
    if len(string) > length:
        i = length - 1
        string = f"{string[:i]}…"
    return string


def unicode_sort(in_list):
    def fmt(string):
        return unidecode.unidecode(string.lower())

    return sorted(in_list, key=fmt)


def sort_xml(xml_tree):
    sort_xml_attributes(xml_tree)
    sort_xml_elements(xml_tree)


def sort_xml_attributes(xml_tree):
    for el in xml_tree.iter():
        attrib = el.attrib
        if len(attrib) > 1:
            attributes = sorted(attrib.items())
            attrib.clear()
            attrib.update(attributes)


def get_sort_value(xml_elem):
    tag = xml_elem.tag
    if isinstance(tag, str):
        return tag
    else:
        # Some tags are Comment objects and can't be sorted as strings.
        return "0"


def sort_xml_elements(xml_tree):
    for el in xml_tree.iter():
        if len(el.getchildren()) > 1:
            sorted_children = sorted(el, key=get_sort_value)
            el[:] = sorted_children


def xmlfile_to_etree(filepath):
    xml_tree = etree.parse(str(filepath), config.XML_PARSER).getroot()
    return xml_tree


def xmlstring_to_etree(xmlstring):
    xml_tree = etree.fromstring(xmlstring, config.XML_PARSER)
    return xml_tree


def etree_to_xmlstring(xml_tree):
    return (
        etree.tostring(
            xml_tree, encoding="UTF-8", pretty_print=True, xml_declaration=True
        )
        .decode()
        .rstrip()
    )


def search_entry(entry, text, field, target_groups, header_fields, match_type, get_all):
    items = []
    if "senses" in target_groups:
        if not entry.sense_items:
            if not get_all:
                return
            else:
                return items
        for sense in entry.sense_items:
            result = search_sense(
                sense, text, field, target_groups, header_fields, match_type, get_all
            )
            if result is not None:
                if not get_all:
                    return result
                else:
                    items.extend(result)

            if not sense.subsense_items:
                continue
            for subsense in sense.subsense_items:
                result = search_sense(
                    subsense,
                    text,
                    field,
                    target_groups,
                    header_fields,
                    match_type,
                    get_all,
                )
                if result is not None:
                    if not get_all:
                        return result
                    else:
                        items.extend(result)

    if "entries" in target_groups:
        if field == "lexical-unit":
            if not entry.lexical_unit or not entry.lexical_unit.form_items:
                if not get_all:
                    return
                else:
                    return items
            if form_items_has_match(entry.lexical_unit.form_items, text, match_type):
                if not get_all:
                    return entry
                else:
                    items.append(entry)
        elif field == "variant":
            if not entry.variant_items:
                if not get_all:
                    return
                else:
                    return items
            for variant in entry.variant_items:
                if not variant.form_items:
                    continue
                if form_items_has_match(variant.form_items, text, match_type):
                    if not get_all:
                        return entry
                    else:
                        items.append(entry)
        elif field in header_fields:
            if not entry.field_items:
                if not get_all:
                    return
                else:
                    return items
            for field_item in entry.field_items:
                if field_item.type == field and field_item.form_items:
                    if form_items_has_match(field_item.form_items, text, match_type):
                        if not get_all:
                            return entry
                        else:
                            items.append(entry)
    if get_all:
        return items


def search_sense(sense, text, field, target_groups, header_fields, match_type, get_all):
    items = []
    if field == "gloss":
        if not sense.gloss_items:
            if not get_all:
                return
            else:
                return items
        for gloss in sense.gloss_items:
            if text in str(gloss.text):
                if not get_all:
                    return sense
                else:
                    items.append(sense)
    elif field == "definition":
        if not sense.definition or not sense.definition.form_items:  # noqa: E501
            if not get_all:
                return
            else:
                return items
        if form_items_has_match(sense.definition.form_items, text, match_type):
            if not get_all:
                return sense
            else:
                items.append(sense)
    elif field == "grammatical-info":
        if not sense.grammatical_info:
            if not get_all:
                return
            else:
                return items
        if text in str(sense.grammatical_info):
            if not get_all:
                return sense
            else:
                items.append(sense)
    elif field in header_fields:
        if not sense.field_items:
            if not get_all:
                return
            else:
                return items
        for field_item in sense.field_items:
            if field_item.type == field and field_item.form_items:
                if form_items_has_match(field_item.form_items, text, match_type):
                    if not get_all:
                        return sense
                    else:
                        items.append(sense)
    if get_all:
        return items


def form_items_has_match(form_items, text, match_type):
    for form in form_items:
        if form.__class__.__name__ == "Span":
            value = ""
            if form.pcdata:
                value += str(form.pcdata)
            if form.tail:
                value += str(form.tail)
        else:
            value = str(form.text)
        if match_type == "contains":
            if text in value:
                return True
        elif match_type == "exact":
            if text == value:
                return True
        elif match_type == "regex":
            if re.match(text, value):
                return True


def get_current_timestamp():
    return datetime.strftime(datetime.now(timezone.utc), "%Y-%m-%dT%H:%M:%SZ")


def get_writing_systems_from_entry(entry) -> dict:
    ws = {"vernacular": [], "analysis": []}
    ws["analysis"].extend(get_ws_from_field_items(entry.field_items))
    if entry.etymology_items:
        for y in entry.etymology_items:
            ws["analysis"].extend(get_ws_from_field_items(y.field_items))
    if entry.lexical_unit and entry.lexical_unit.form_items:
        for f in entry.lexical_unit.form_items:
            ws["vernacular"].append(f.lang)
    if entry.note_items:
        for n in entry.note_items:
            ws["analysis"].extend(get_ws_from_field_items(n.field_items))
    if entry.pronunciation_items:
        for p in entry.pronunciation_items:
            if p.form_items:
                for f in p.form_items:
                    ws["vernacular"].append(f.lang)
            ws["analysis"].extend(get_ws_from_field_items(p.field_items))
    if entry.variant_items:
        for v in entry.variant_items:
            ws["analysis"].extend(get_ws_from_field_items(v.field_items))
    if entry.sense_items:
        for s in entry.sense_items:
            ws["analysis"].extend(get_ws_from_sense(s))

    # Remove repeated elements.
    ws["vernacular"] = list(set(ws.get("vernacular")))
    ws["analysis"] = list(set(ws.get("analysis")))
    return ws


def get_ws_from_sense(sense) -> list:
    analysis_ws = []
    analysis_ws.extend(get_ws_from_field_items(sense.field_items))
    if sense.example_items:
        for e in sense.example_items:
            analysis_ws.extend(get_ws_from_field_items(e.field_items))
    if sense.note_items:
        for n in sense.note_items:
            analysis_ws.extend(get_ws_from_field_items(n.field_items))
    if sense.relation_items:
        for r in sense.relation_items:
            analysis_ws.extend(get_ws_from_field_items(r.field_items))
    return list(set(analysis_ws))


def get_ws_from_field_items(field_items):
    writing_systems = []
    if field_items is not None:
        for field in field_items:
            if field.form_items:
                for m in field.form_items:
                    writing_systems.append(m.lang)
    return writing_systems



import unittest

from lxml import etree

from lift_utils import config
from lift_utils.lexicon import Lexicon
from lift_utils.utils import xmlfile_to_etree

from . import DATA_PATH, SANGO_LIFT

LIFT_GOOD = SANGO_LIFT
LEXICON = Lexicon(LIFT_GOOD)


class TestFind(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON
        config.LIFT_VERSION = "0.13"  # insurance

    def test_find_all_definition(self):
        self.assertEqual(len(self.lexicon.find_all("husband", field="definition")), 2)

    def test_find_all_gloss(self):
        self.assertEqual(len(self.lexicon.find_all("head")), 20)

    def test_find_all_grammatical_info(self):
        self.assertEqual(
            len(self.lexicon.find_all("Nom", field="grammatical-info")), 2381
        )

    def test_find_all_header_field_contains(self):
        self.assertEqual(len(self.lexicon.find_all("1021", field="CAWL")), 1)

    def test_find_all_header_field_exact(self):
        self.assertEqual(
            len(self.lexicon.find_all("1056", field="CAWL", match_type="exact")), 1
        )

    def test_find_definition_contains(self):
        self.assertEqual(
            str(self.lexicon.find("husbandry", field="definition").id),
            # "f6c860ea-80ff-431a-a5dd-9469b315f40c",  # id for "man"; wrong?
            "ebcf4d79-6bb9-4d33-9f31-cdc03355a718",  # id for "husbandry"
        )

    def test_find_gloss_contains(self):
        self.assertEqual(
            str(self.lexicon.find("house").id), "086e06e3-a5f6-4c54-805a-4a1adcd51d4f"
        )

    def test_find_header_field_contains(self):
        self.assertEqual(
            str(self.lexicon.find("i-1056", field="CAWL").id),
            "25222115-e109-4704-8e82-d7cd523c8cbd",
        )

    def test_find_header_field_exact(self):
        self.assertIsNone(self.lexicon.find("056", field="CAWL", match_type="exact"))


class TestParse(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON
        config.LIFT_VERSION = "0.13"  # insurance

    def test_lexicon_entry_items(self):
        self.assertTrue(len(self.lexicon.entry_items) == 3507)

    def test_lexicon_find_writing_systems(self):
        self.lexicon.vernacular_writing_systems.sort()
        self.lexicon.analysis_writing_systems.sort()
        self.assertEqual(self.lexicon.vernacular_writing_systems, ["sg", "sg-fonipa"])
        self.assertEqual(self.lexicon.analysis_writing_systems, ["de", "en", "fr"])

    def test_lexicon_get_item_from_id(self):
        refid = "ndâmbo kôlï/wâlï_fff4e6bb-9c09-43d6-ac11-b9220141b52b"
        item = self.lexicon.get_item_by_id(refid)
        self.assertIsNotNone(item)
        self.assertEqual(item.date_modified, "2023-04-20T19:31:03Z")

    # def test_lexicon_get_parent_from_id(self):
    #     refid = "25222115-e109-4704-8e82-d7cd523c8cbd"
    #     parent = self.lexicon.get_item_parent_by_id(refid)
    #     self.assertIsNotNone(parent)
    #     self.assertEqual(parent.date_modified, "2023-04-20T19:31:03Z")

    def test_lexicon_header_field_items(self):
        self.assertTrue(len(self.lexicon.header.fields.field_items) == 8)

    def test_lexicon_header_ranges(self):
        self.assertTrue(len(self.lexicon.header.ranges.range_items) == 20)


class TestReadWrite(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON
        config.LIFT_VERSION = "0.13"  # insurance

    # NOTE: I haven't yet found a reliable way to compare XML file data. Using
    # "canonicalize" and "tostring" both result in XML strings whose child
    # element orders depend on the tree order rather than some kind of sorted
    # order useful for comparison. And since the element order of the input
    # file is unpredictable, there's no way to guarantee that the output file
    # will produce the same order, even if contents are identical.
    def test_read_write_file(self):
        infile = LIFT_GOOD
        outfile = DATA_PATH / "sango_out.lift"
        self.lexicon.to_lift(outfile)
        xml_in = (
            etree.tostring(
                xmlfile_to_etree(infile),
                encoding="UTF-8",
                pretty_print=True,
                xml_declaration=True,
            )
            .decode()
            .rstrip()
        )
        xml_out = (
            etree.tostring(
                xmlfile_to_etree(outfile),
                encoding="UTF-8",
                pretty_print=True,
                xml_declaration=True,
            )
            .decode()
            .rstrip()
        )

        ranges_xml_in = (
            etree.tostring(
                xmlfile_to_etree(infile.with_suffix(".lift-ranges")),
                encoding="UTF-8",
                pretty_print=True,
                xml_declaration=True,
            )
            .decode()
            .rstrip()
        )
        ranges_xml_out = (
            etree.tostring(
                xmlfile_to_etree(outfile.with_suffix(".lift-ranges")),
                encoding="UTF-8",
                pretty_print=True,
                xml_declaration=True,
            )
            .decode()
            .rstrip()
        )
        # xml_in = etree.canonicalize(
        #     xmlfile_to_etree(infile),
        #     strip_text=True
        # )
        # xml_out = etree.canonicalize(
        #     xmlfile_to_etree(outfile),
        #     strip_text=True
        # )
        # self.maxDiff = None
        # NOTE: In lieu of being able to compare xml content, just using a file
        # size comparison for the time being. This works okay as long as there
        # aren't too many comments in the infile, since comments aren't kept.
        # This tests that the outfile's size is within the given % of the infile's.
        delta = 0.05
        # print(len(xml_out)/len(xml_in))
        # print(len(ranges_xml_out)/len(ranges_xml_in))
        self.assertAlmostEqual(1, len(xml_out) / len(xml_in), delta=delta)
        self.assertAlmostEqual(1, len(ranges_xml_out) / len(ranges_xml_in), delta=delta)



class InvalidExtensionError(Exception):
    def __init__(self, filename):
        message = f"File has an invalid extension: {filename}"
        super().__init__(message)


class RequiredValueError(Exception):
    def __init__(self, values):
        message = f"Required value(s) missing: {', '.join(values)}"
        super().__init__(message)


class UnsupportedActionException(Exception):
    def __init__(self, version):
        message = f"Action not supported in LIFT v{version}"
        super().__init__(message)



"""Manipulate base linguistic elements."""

import sys
from typing import List, Optional

from lxml import etree

from . import config
from .datatypes import URL, DateTime, Key, Lang, PCData
from .errors import RequiredValueError
from .utils import etree_to_xmlstring


class LIFTUtilsBase:
    """This is a base class for all LIFT nodes.

    :ivar etree._Element xml_tree: The node's current data.
    """

    XML_TAG = None
    XML_PY_NAMES = {
        # keys are LIFT XML tags; values are Python object properties
        "abbrev": "abbrev_items",
        "annotation": "annotation_items",
        "class": "class_",
        "dateCreated": "date_created",
        "dateDeleted": "date_deleted",
        "dateModified": "date_modified",
        "description": "description_items",
        "entry": "entry_items",
        "etymology": "etymology_items",
        "example": "example_items",
        "field": "field_items",
        "form": "form_items",
        "gloss": "gloss_items",
        "grammatical-info": "grammatical_info",
        "illustration": "illustration_items",
        "label": "label_items",
        "lexical-unit": "lexical_unit",
        "media": "media_items",
        "note": "note_items",
        "option-range": "option_range",
        "pronunciation": "pronunciation_items",
        "range-element": "range_element_items",
        "range": "range_items",
        "relation": "relation_items",
        "reversal": "reversal_items",
        "sense": "sense_items",
        "span": "span_items",
        "subsense": "subsense_items",
        "trait": "trait_items",
        "translation": "translation_items",
        "usage": "usage_items",
        "variant": "variant_items",
        "writing-system": "writing_system",
    }

    def __init__(self, parent_item=None, xml_tree: etree._Element = None):
        # Initialize property values.
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set()
        # Store a cumulative dict of key="tag name," value="Python class". This
        # can't be defined declaratively because classes must be defined first,
        # but this also needs to be referred to before all classes are defined.
        config.TAG_CLASSES.update({"tail": PCData})
        # Link to parent node for tree traversal.
        self.parent_item = parent_item

    def print(self, _format="xml"):
        """Print the node's data to stdout; as XML by default."""
        try:
            if _format == "xml":
                print(self._to_xml(), flush=True)
            else:
                raise NotImplementedError
        except BrokenPipeError:
            sys.stdout = None

    def prop_name_from_xml_name(self, xml_name):
        """Return the object property name from the given XML attribute or
        element name."""

        name = None
        exc = {
            "description": (
                "FieldDefinition",
                "Header",
                "Range",
                "Range13",
            ),
            "form": ("Etymology",),
            "label": (
                "FieldDefinition",
                "URLRef",
            ),
        }
        for x_name, classes in exc.items():
            if self.__class__.__name__ in classes and xml_name == x_name:
                # Exceptional situations where "description" is the python attrib
                # rather than "description_items", so the XML_PY_NAMES conversion
                # dict should not be used.
                name = xml_name
        if name is None:
            name = self.XML_PY_NAMES.get(xml_name, xml_name)
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            if self.__class__.__name__ == "Field" and name == "name":
                name = "type"
        else:
            if self.__class__.__name__ == "Field" and name == "type":
                name = "name"
        return name

    def show(self):
        """Print an overview of the object in the terminal window."""
        for k, v in self.__dict__.items():
            if not k.startswith("_"):
                print(f"{k}: {v}")

    def _add_list_item(self, _name, _class, **kwargs):
        new_obj = _class(parent_item=self, **kwargs)
        if getattr(self, _name) is None:
            setattr(self, _name, [new_obj])
        else:
            getattr(self, _name).append(new_obj)
        return new_obj

    def _from_xml_tree(self, xml_tree):
        # Convert XML attributes to python properties.
        attribs = [a for a in self._attributes_required]
        attribs.extend([a for a in self._attributes_optional])
        for xml_name in attribs:
            py_name = self.prop_name_from_xml_name(xml_name)
            py_cls = config.TAG_CLASSES.get(xml_name)
            if xml_name in xml_tree.attrib.keys():
                setattr(
                    self,
                    py_name,
                    py_cls(xml_tree.attrib.get(xml_name)),
                )

        # Convert properties to XML elements.
        elems = [e for e in self._elements_required]
        elems.extend([e for e in self._elements_optional])
        for xml_name in elems:
            py_name = self.prop_name_from_xml_name(xml_name)
            py_cls = config.TAG_CLASSES.get(xml_name)
            if xml_name == "pcdata":
                if xml_tree.text:
                    setattr(self, py_name, py_cls(xml_tree.text))
                continue
            elif xml_name == "tail":
                if xml_tree.tail:
                    setattr(self, py_name, py_cls(xml_tree.tail))
                continue

            # Handle child elements.
            for c in xml_tree.getchildren():
                if c.tag == xml_name:
                    if py_name.endswith("_items"):  # list-like obj/elem
                        if not getattr(self, py_name):
                            # Instantiate list-like object.
                            setattr(self, py_name, list())
                        getattr(self, py_name).append(
                            py_cls(xml_tree=c, parent_item=self)
                        )
                    else:  # single element
                        setattr(self, py_name, py_cls(xml_tree=c, parent_item=self))

    def _to_xml_tree(self, tag=None):
        # TODO: Why isn't self.XML_TAG sufficient in every case here?
        if tag:
            xml_tag = tag
        else:
            xml_tag = self.XML_TAG
        xml_tree = etree.Element(xml_tag)

        # Convert properties to XML attributes.
        attribs = [a for a in self._attributes_required]
        attribs.extend([a for a in self._attributes_optional])
        for xml_name in attribs:
            # Get attribute's value from attribute's XML name, which might be
            # different from the Python name.
            val = getattr(self, self.prop_name_from_xml_name(xml_name))
            if val is not None:
                xml_tree.set(xml_name, str(val))

        # Convert properties to XML elements.
        elems = [e for e in self._elements_required]
        elems.extend([e for e in self._elements_optional])
        for xml_name in elems:
            py_name = self.prop_name_from_xml_name(xml_name)
            val = getattr(self, py_name)
            if not val:
                continue
            if hasattr(val, "append"):  # list-like child element
                for o in getattr(self, py_name):
                    xml_tree.append(o._to_xml_tree(tag=xml_name))
            elif py_name == "pcdata":  # special text element
                xml_tree.text = val
            elif py_name == "tail":  # special tail element
                xml_tree.tail = val
            else:  # single child element
                o = getattr(self, py_name)
                xml_tree.append(o._to_xml_tree(tag=xml_name))
        return xml_tree

    def _to_xml(self, xml_tree=None):
        if xml_tree is None:
            xml_tree = self._to_xml_tree()
        return etree_to_xmlstring(xml_tree)


class Span(LIFTUtilsBase):
    """A Unicode string marked with language and formatting information."""

    XML_TAG = "span"

    def __init__(
        self,
        text: str = None,
        tail: str = None,
        lang: str = None,
        href: str = None,
        span_class: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(("class", "href", "lang"))
        self._elements_required = set(("pcdata",))
        self._elements_optional = set(("span", "tail"))
        config.TAG_CLASSES.update(
            {
                "lang": Lang,
                "href": URL,
                "class": str,
                "pcdata": PCData,
                "span": Span,
            }
        )

        # attributes
        self.lang: Optional[Lang] = None
        self.href: Optional[URL] = None
        self.class_: Optional[str] = None
        # elements
        self.pcdata: PCData = None
        self.tail: Optional[PCData] = None
        self.span_items: Optional[List[Span]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        elif text is not None:
            self.pcdata = PCData(text)
            if tail is not None:
                self.tail = PCData(tail)
            if lang is not None:
                self.lang = Lang(lang)
            if href is not None:
                self.href = URL(href)
            if span_class is not None:
                self.class_ = span_class
        else:
            raise RequiredValueError(("text",))

    def __str__(self):
        s = self.pcdata
        if self.span_items:
            for item in self.span_items:
                s += item.pcdata
                if item.tail:
                    s += item.tail
        if self.tail:
            s += self.tail


class Trait(LIFTUtilsBase):
    """An important mechanism for giving type information to an object.
    It can also be used for adding binary constraints.
    """

    XML_TAG = "trait"

    def __init__(
        self,
        name: str = None,
        value: str = None,
        trait_id: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("name", "value"))
        self._attributes_optional = set(("id",))
        self._elements_required = set()
        self._elements_optional = set(("annotation",))
        config.TAG_CLASSES.update(
            {
                "name": Key,
                "value": Key,
                "id": Key,
                "annotation": Annotation,
            }
        )

        # attributes
        self.name: Key = None
        self.value: Key = None
        self.id: Optional[Key] = None
        # elements
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        elif name is not None and value is not None:
            self.name = Key(name)
            self.value = Key(value)
            if trait_id is not None:
                self.id = Key(trait_id)
        else:
            raise RequiredValueError(("name", "value"))

    def __str__(self):
        return f"{self.name}: {self.value}"


class Flag(Trait):
    """An important mechanism for giving type information to an object.
    It can also be use for adding binary constraints.

    .. note:: Used by LIFT v0.13 (FieldWorks). Mentioned in specification but
        not defined; assumed to be equivalent to ``Trait``.
    """

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)


class Text(LIFTUtilsBase):
    """Contains textual data mixed with ``span`` elements only."""

    XML_TAG = "text"

    def __init__(
        self,
        text: str = None,
        xml_tree: Optional[etree._Element] = None,
        subinit=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set(("pcdata",))
        self._elements_optional = set(("span",))
        config.TAG_CLASSES.update(
            {
                "pcdata": PCData,
                "span": Span,
            }
        )

        # elements
        self.pcdata: PCData = None
        self.span_items: Optional[List[Span]] = None

        if xml_tree is not None:
            if xml_tree is False:
                return
            self._from_xml_tree(xml_tree)
        elif text is not None:
            self.pcdata = PCData(text)
        else:
            raise RequiredValueError(("text",))

    def __str__(self):
        if self.span_items:
            s = f"{str(self.pcdata)}{''.join(str(s) for s in self.span_items)}"
        else:
            s = str(self.pcdata)
        return s


class Form(LIFTUtilsBase):
    """A representation of a string in a particular language and script.
    This is specified by the ``lang`` attribute.
    """

    XML_TAG = "form"

    def __init__(
        self,
        lang: Lang = None,
        text: Text = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("lang",))
        self._attributes_optional = set()
        self._elements_required = set(("text",))
        self._elements_optional = set(("annotation",))
        config.TAG_CLASSES.update(
            {
                "lang": Lang,
                "text": Text,
                "annotation": Annotation,
            }
        )

        # attributes
        self.lang: Lang = None
        # elements
        self.text: Text = None
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            if xml_tree is False:
                return
            self._from_xml_tree(xml_tree)
        elif lang is not None and text is not None:
            self.lang = Lang(lang)
            self.text = Text(text=text)
        else:
            raise RequiredValueError(("lang", "text"))

    def __str__(self):
        return str(self.text)


class URLRef(LIFTUtilsBase):
    """This is a URL with a caption."""

    XML_TAG = "urlref"

    def __init__(
        self,
        href: str = None,
        label: dict = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("href",))
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set(("label",))
        config.TAG_CLASSES.update(
            {
                "href": URL,
                "label": Multitext,
            }
        )

        # attributes
        self.href: URL = None
        # elements
        # NOTE: `label` elements are not explicitly defined as possibly having
        # multple values here, but they are for LiftRanges and RangeElements.
        self.label: Optional[Multitext] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        elif href is not None:
            self.href = URL(href)
            if label is not None:
                self.set_label(label)
        else:
            raise RequiredValueError(("href",))

    def set_label(self, label_dict):
        if not isinstance(label_dict, dict):
            raise RequiredValueError(("dict of {{lang: text}} pairs",))
        self.label = Multitext(label_dict)

    def __str__(self):
        return str(self.href)


class Multitext(Text):
    """Allows for different representations of the same information.
    It can be in a given language, or in multiple languages.
    """

    # Multitext is only used as a super class for other classes, so it
    # doesn't have it's own XML tag.
    XML_TAG = None

    def __init__(
        self,
        form_dict: dict = None,
        # trait_dict: dict = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(
            xml_tree=False,
            **kwargs,
        )
        self._attributes_required = set()
        self._attributes_optional = set()
        self._elements_required = set()
        self._elements_optional = set(("form", "pcdata", "span", "trait"))
        config.TAG_CLASSES.update(
            {
                "form": Form,
                "trait": Trait,
            }
        )

        # elements
        self.form_items: Optional[List[Form]] = None
        self.trait_items: Optional[List[Trait]] = None

        if xml_tree is not None:
            if xml_tree is False:
                return
            self._from_xml_tree(xml_tree)
        elif form_dict is not None:
            self.set_form_items(form_dict)

    def __str__(self):
        s = "multitext"
        if self.form_items:
            # self.form_items.sort()
            s = f"{self.form_items[0].text} ({self.form_items[0].lang})"
            ct = len(self.form_items)
            if ct > 1:
                s = f"{s} ({ct} forms)"
        return s

    def get_form_by_lang(self, lang):
        """Return first form item with matching ``lang`` attribute.

        :var str lang: Language code; e.g. "en".
        """
        for f in self.form_items:
            if f.lang == lang:
                return f

    def set_form_items(self, form_dict):
        self.form_items = []
        for lg, tx in form_dict.items():
            self.form_items.append(Form(lang=lg, text=tx))


class Gloss(Form):
    """A representation of a sense's gloss.
    It's given in a particular language and script as specified by the ``lang``
    attribute.
    """

    XML_TAG = "gloss"

    def __init__(
        self,
        lang: str = None,
        text: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(xml_tree=False, **kwargs)
        self._attributes_required = set(("lang",))
        self._attributes_optional = set()
        self._elements_required = set(("text",))
        self._elements_optional = set(("annotation", "trait"))
        config.TAG_CLASSES.update(
            {
                "trait": Trait,
            }
        )

        # elements
        self.trait_items: Optional[List[Trait]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        elif lang is not None and text is not None:
            self.lang = Lang(lang)
            self.text = Text(text=text)
        else:
            raise RequiredValueError(("lang", "text"))

    def __str__(self):
        return f"{self.text} ({self.lang})"

    def set_trait_items(self, trait_dict):
        self.trait_items = []
        for name, data in trait_dict.items():
            value = data.get("value")
            trait_id = data.get("trait_id")
            self.trait_items.append(Trait(name=name, value=value, trait_id=trait_id))


class Annotation(Multitext):
    """Provides a mechanism for adding meta-information to almost any element."""

    XML_TAG = "annotation"

    def __init__(
        self,
        name: str = None,
        value: str = None,
        who: str = None,
        when: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("name", "value"))
        self._attributes_optional = set(("when", "who"))
        self._elements_required = set()
        self._elements_optional = set(("form", "pcdata", "span", "trait"))
        config.TAG_CLASSES.update(
            {
                "name": Key,
                "value": Key,
                "who": Key,
                "when": DateTime,
            }
        )

        # attributes
        self.name: Key = None
        self.value: Key = None
        self.who: Optional[Key] = None
        self.when: Optional[DateTime] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)
        elif name is not None and value is not None:
            self.name = Key(name)
            self.value = Key(value)
            if who is not None:
                self.who = Key(who)
            if when is not None:
                self.when = DateTime(when)
        else:
            raise RequiredValueError(("name", "value"))


class Field(Multitext):
    """A generalised element to allow an application to store information.
    It's for information in a LIFT file that isn't explicitly described in the
    LIFT standard.
    """

    XML_TAG = "field"

    def __init__(
        self,
        field_type: str = None,
        name: str = None,
        xml_tree: Optional[etree._Element] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._attributes_required = set(("name",))
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            self._attributes_required = set(("type",))
        self._attributes_optional = set(("dateCreated", "dateModified"))
        self._elements_required = set()
        self._elements_optional = set(("annotation", "form", "pcdata", "span", "trait"))
        config.TAG_CLASSES.update(
            {
                "type": Key,
                "name": Key,
                "dateCreated": DateTime,
                "dateModified": DateTime,
                "annotation": Annotation,
                "trait": Trait,
            }
        )

        # attributes
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            self.type: Key = None
        else:
            self.name: Key = None
        self.date_created: Optional[DateTime] = None
        self.date_modified: Optional[DateTime] = None
        # elements
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            # self.trait_items: Optional[List[Flag]] = None
            self.trait_items: Optional[List[Trait]] = None
            # NOTE: I think there must be a typo in the docs. I think the
            # 'form_items' class should be "Form" rather than "Span", and "Form"
            # is already accounted for in the subclassing of 'Multitext'.
            # self.form_items: Optional[List[Span]] = None
        else:
            self.trait_items: Optional[List[Trait]] = None
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

        missing_required_attrib = None
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            if field_type is not None:
                self.type = Key(field_type)
            elif xml_tree is None:
                missing_required_attrib = "field_type"
        else:
            if name is not None:
                self.name = Key(name)
            elif xml_tree is None:
                missing_required_attrib = "name"

        if missing_required_attrib:
            raise RequiredValueError((missing_required_attrib,))


class Extensible(LIFTUtilsBase):
    """Used to provide certain extra information in a controlled way.

    :ivar Optional[DateTime] date_created: Contains a date/timestamp saying
        when the element was added to the dictionary.
    :ivar Optional[DateTime] date_modified: Contains a date/timestamp saying
        when the element was last changed.
    :ivar Optional[List[Field]] field_items: Holds extra textual information.
    :ivar Optional[List[Trait]] trait_items: Adds type or constraint
        information.
    :ivar Optional[List[Annotation]] annotation: Adds meta-information
        describing the element.
    """

    # Extensible is only used as a super class for other classes, so it
    # doesn't need its own XML tag.
    XML_TAG = None

    def __init__(self, xml_tree: Optional[etree._Element] = None, **kwargs):
        super().__init__(**kwargs)
        self._attributes_required = set()
        self._attributes_optional = set(("dateCreated", "dateModified"))
        self._elements_required = set()
        self._elements_optional = set(("annotation", "field", "trait"))
        config.TAG_CLASSES.update(
            {
                "dateCreated": DateTime,
                "dateModified": DateTime,
                "field": Field,
                "trait": Trait,
                "annotation": Annotation,
            }
        )

        # attributes
        self.date_created: Optional[DateTime] = None
        self.date_modified: Optional[DateTime] = None
        # elements
        self.field_items: Optional[List[Field]] = None
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            # TODO: Find definition of "Flag" in v0.13?
            # self.trait_items: Optional[List[Flag]] = None
            self.trait_items: Optional[List[Trait]] = None
        else:
            self.trait_items: Optional[List[Trait]] = None
        self.annotation_items: Optional[List[Annotation]] = None

        if xml_tree is not None:
            self._from_xml_tree(xml_tree)

    def add_annotation(self, name=None, value=None, who=None, when=None):
        return self._add_list_item(
            "annotation_items",
            Annotation,
            name=name,
            value=value,
            who=who,
            when=when,
        )

    def add_field(self, name=None):
        if config.LIFT_VERSION == config.LIFT_VERSION_FIELDWORKS:
            kwargs = {"field_type": name}
        else:
            kwargs = {"name": name}
        return self._add_list_item("field_items", Field, **kwargs)

    def add_trait(self, name=None, value=None, trait_id=None):
        return self._add_list_item(
            "trait_items",
            Trait,
            name=name,
            value=value,
            trait_id=trait_id,
        )

    def set_date_created(self):
        self.date_created = DateTime()

    def set_date_modified(self):
        self.date_modified = DateTime()



import unittest

from lift_utils import base, datatypes, errors, lexicon

from . import SANGO_LIFT

# NOTE: Loading the LEXICON one time because of heavy CPU overhead, and because
# modfied times will change during testing.
LEXICON = lexicon.Lexicon(SANGO_LIFT)
FIRST_ENTRY = LEXICON.entry_items[0]
FIRST_ENTRY_LAST_MODIFIED = FIRST_ENTRY.date_modified
FIRST_SENSE = FIRST_ENTRY.sense_items[0]
FIRST_SENSE_LAST_MODIFIED = FIRST_SENSE.date_modified


class TestAddExtensibleItems(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON
        self.entry = self.lexicon.entry_items[0]

    def test_add_annotation(self):
        self.assertRaises(
            errors.RequiredValueError, self.entry.add_annotation, name="name"
        )
        idx = len(self.entry.annotation_items) if self.entry.annotation_items else 0
        self.entry.add_annotation(name="test", value="test value")
        self.assertIsInstance(self.entry.annotation_items[idx], base.Annotation)
        self.assertEqual(len(self.entry.annotation_items), 1)
        del self.entry.annotation_items[idx]

    def test_add_field(self):
        self.assertRaises(errors.RequiredValueError, self.entry.add_field)
        idx = len(self.entry.field_items) if self.entry.field_items else 0
        self.entry.add_field(name="Test Field")
        self.assertIsInstance(self.entry.field_items[idx], base.Field)
        self.assertEqual(self.entry.field_items[idx].type, "Test Field")
        del self.entry.field_items[idx]

    def test_add_trait(self):
        self.assertRaises(errors.RequiredValueError, self.entry.add_trait)
        idx = len(self.entry.trait_items) if self.entry.trait_items else 0
        self.entry.add_trait(name="trait-name", value="trait-value")
        self.assertIsInstance(self.entry.trait_items[idx], base.Trait)
        self.assertEqual(self.entry.trait_items[idx].name, "trait-name")
        self.assertEqual(self.entry.trait_items[idx].value, "trait-value")
        del self.entry.trait_items[idx]


class TestCreateBaseItems(unittest.TestCase):
    def test_create_annotation_missing_info(self):
        self.assertRaises(errors.RequiredValueError, base.Annotation)
        self.assertRaises(errors.RequiredValueError, base.Annotation, name="test")

    def test_create_annotation_optional(self):
        self.anno = base.Annotation(
            name="optional-name",
            value="optional-value",
            who="optional-person",
            when="optional-timestamp",
        )
        self.assertIsInstance(self.anno.who, datatypes.Key)
        self.assertIsInstance(self.anno.when, datatypes.DateTime)
        self.assertEqual(self.anno.who, "optional-person")
        self.assertEqual(self.anno.when, "optional-timestamp")

    def test_create_annotation_required(self):
        self.anno = base.Annotation(name="test", value="value")
        self.assertIsInstance(self.anno.name, datatypes.Key)
        self.assertIsInstance(self.anno.value, datatypes.Key)
        self.assertEqual(self.anno.name, "test")
        self.assertEqual(self.anno.value, "value")

    def test_create_annotation(self):
        self.assertRaises(errors.RequiredValueError, base.Annotation)
        self.assertRaises(errors.RequiredValueError, base.Annotation, name="test")
        self.anno = base.Annotation(
            name="name", value="value", who="optional-person", when="optional-timestamp"
        )
        self.assertIsInstance(self.anno.name, datatypes.Key)
        self.assertIsInstance(self.anno.value, datatypes.Key)
        self.assertIsInstance(self.anno.who, datatypes.Key)
        self.assertIsInstance(self.anno.when, datatypes.DateTime)
        self.assertEqual(self.anno.name, "name")
        self.assertEqual(self.anno.value, "value")
        self.assertEqual(self.anno.who, "optional-person")
        self.assertEqual(self.anno.when, "optional-timestamp")

    def test_create_field(self):
        self.assertRaises(errors.RequiredValueError, base.Field)
        self.field = base.Field(field_type="test-field")
        self.assertIsInstance(self.field.type, datatypes.Key)
        self.assertEqual(self.field.type, "test-field")

    def test_create_form(self):
        self.assertRaises(errors.RequiredValueError, base.Form)
        self.form = base.Form(lang="en", text="text")
        self.assertIsInstance(self.form.lang, datatypes.Lang)
        self.assertIsInstance(self.form.text, base.Text)
        self.assertEqual(self.form.lang, "en")
        self.assertEqual(str(self.form.text), "text")

    def test_create_span(self):
        self.assertRaises(errors.RequiredValueError, base.Span)
        self.span = base.Span(
            text="this text",
            lang="en",
            href="https://example.com",
            span_class="bold",
            tail="tail text",
        )
        self.assertIsInstance(self.span.pcdata, datatypes.PCData)
        self.assertIsInstance(self.span.tail, datatypes.PCData)
        self.assertIsInstance(self.span.lang, datatypes.Lang)
        self.assertIsInstance(self.span.href, datatypes.URL)
        self.assertIsInstance(self.span.class_, str)
        self.assertEqual(self.span.pcdata, "this text")
        self.assertEqual(self.span.tail, "tail text")
        self.assertEqual(self.span.lang, "en")
        self.assertEqual(self.span.href, "https://example.com")
        self.assertEqual(self.span.class_, "bold")

    def test_create_text(self):
        self.assertRaises(errors.RequiredValueError, base.Text)
        self.text = base.Text(text="text")
        self.assertIsInstance(self.text.pcdata, datatypes.PCData)
        self.assertEqual(self.text.pcdata, "text")

    def test_create_trait(self):
        self.assertRaises(errors.RequiredValueError, base.Trait)
        self.trait = base.Trait(name="name", value="value", trait_id="trait-id")
        self.assertIsInstance(self.trait.name, datatypes.Key)
        self.assertIsInstance(self.trait.value, datatypes.Key)
        self.assertIsInstance(self.trait.id, datatypes.Key)
        self.assertEqual(self.trait.name, "name")
        self.assertEqual(self.trait.value, "value")
        self.assertEqual(self.trait.id, "trait-id")

    def test_create_urlref(self):
        self.assertRaises(errors.RequiredValueError, base.URLRef)
        self.url = base.URLRef(href="https://www.example.com")
        self.assertIsInstance(self.url.href, datatypes.URL)
        self.assertEqual(self.url.href, "https://www.example.com")


class TestModifyLexiconItems(unittest.TestCase):
    def setUp(self):
        self.lexicon = LEXICON
        self.entry = FIRST_ENTRY
        self.sense = FIRST_SENSE
        self.entry_last_modified = FIRST_ENTRY_LAST_MODIFIED
        self.sense_last_modified = FIRST_SENSE_LAST_MODIFIED

    def test_entry_add_etymology(self):
        idx = len(self.entry.etymology_items) if self.entry.etymology_items else 0
        self.entry.add_etymology()
        self.assertIsInstance(idx, int)
        self.assertIsInstance(self.entry.etymology_items[idx], lexicon.Etymology)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)
        del self.entry.etymology_items[idx]

    def test_entry_add_note(self):
        idx = len(self.entry.note_items) if self.entry.note_items else 0
        self.entry.add_note()
        self.assertIsInstance(idx, int)
        self.assertIsInstance(self.entry.note_items[idx], lexicon.Note)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)
        del self.entry.note_items[idx]

    def test_entry_add_pronunciation(self):
        idx = (
            len(self.entry.pronunciation_items) if self.entry.pronunciation_items else 0
        )
        self.entry.add_pronunciation()
        self.assertIsInstance(idx, int)
        self.assertIsInstance(self.entry.pronunciation_items[idx], lexicon.Phonetic)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)
        del self.entry.pronunciation_items[idx]

    def test_entry_add_relation(self):
        idx = len(self.entry.relation_items) if self.entry.relation_items else 0
        self.entry.add_relation()
        self.assertIsInstance(idx, int)
        self.assertIsInstance(self.entry.relation_items[idx], lexicon.Relation)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)
        del self.entry.relation_items[idx]

    def test_entry_add_sense(self):
        idx = len(self.entry.sense_items) if self.entry.sense_items else 0
        self.entry.add_sense()
        self.assertIsInstance(idx, int)
        self.assertIsInstance(self.entry.sense_items[idx], lexicon.Sense)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)
        del self.entry.sense_items[idx]

    def test_entry_add_variant(self):
        idx = len(self.entry.variant_items) if self.entry.variant_items else 0
        self.entry.add_variant()
        self.assertIsInstance(idx, int)
        self.assertIsInstance(self.entry.variant_items[idx], lexicon.Variant)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)
        del self.entry.variant_items[idx]

    def test_entry_set_lexical_unit(self):
        self.entry.set_lexical_unit({"en": "english text", "sg": "atëne ti sängö"})
        self.assertEqual(len(self.entry.lexical_unit.form_items), 2)
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)

    def test_entry_set_pronunciation_item(self):
        entry_id = "rô_ff402c0b-4eb5-4df1-bd29-a1900aaf7567"
        entry = self.lexicon.get_item_by_id(entry_id)
        labels = {
            "en": "pronunciation recording",
            "fr": "enregistrement de prononciation",
        }
        entry.pronunciation_items[0].add_media(
            href="file:///home/user/pronunciation.wav", label=labels
        )
        media_item = entry.pronunciation_items[0].media_items[0]
        self.assertEqual(media_item.href, "file:///home/user/pronunciation.wav")
        self.assertEqual(len(media_item.label.form_items), 2)
        for f in media_item.label.form_items:
            self.assertIn(f.lang, labels.keys())
            self.assertIn(str(f.text), labels.values())
        self.assertIsNotNone(self.entry.date_modified)
        self.assertNotEqual(self.entry.date_modified, self.entry_last_modified)

    def test_etymology_add_gloss_item(self):
        entry_id = "yongo_eafee046-6df9-4162-8c38-a2536713b69f"
        entry = self.lexicon.get_item_by_id(entry_id)
        entry.etymology_items[0].add_gloss("en", "English gloss")
        idx = len(entry.etymology_items[0].gloss_items) - 1
        gloss_item = entry.etymology_items[0].gloss_items[idx]
        self.assertEqual(gloss_item.lang, "en")
        self.assertEqual(str(gloss_item.text), "English gloss")
        del entry.etymology_items[0].gloss_items[idx]

    def test_lexicon_add_entry_item(self):
        len_before = len(self.lexicon.entry_items)
        self.lexicon.add_entry()
        len_after = len(self.lexicon.entry_items)
        idx = len_after - 1

        self.assertEqual(len_after - 1, len_before)
        self.assertIsNotNone(self.lexicon.entry_items[idx].id)
        self.assertIsNotNone(self.lexicon.entry_items[idx].date_created)
        del self.lexicon.entry_items[idx]

    def test_sense_add_example(self):
        idx = len(self.sense.example_items) if self.sense.example_items else 0
        self.sense.add_example()
        self.assertEqual(len(self.sense.example_items), 1)
        self.assertIsInstance(self.sense.example_items[idx], lexicon.Example)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.example_items[idx]

    def test_sense_add_gloss(self):
        idx = len(self.sense.gloss_items) if self.sense.gloss_items else 0
        self.sense.add_gloss(lang="en", text="gloss text")
        len_after = len(self.sense.gloss_items)
        self.assertEqual(len_after - 1, idx)
        self.assertIsInstance(self.sense.gloss_items[idx], base.Gloss)
        self.assertEqual(self.sense.gloss_items[idx].lang, "en")
        self.assertEqual(str(self.sense.gloss_items[idx].text), "gloss text")
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.gloss_items[idx]

    def test_sense_add_illustration(self):
        href = "file:///home/user/image.png"
        idx = len(self.sense.illustration_items) if self.sense.illustration_items else 0
        self.sense.add_illustration(href=href)
        self.assertEqual(len(self.sense.illustration_items), 1)
        self.assertIsInstance(self.sense.illustration_items[idx], base.URLRef)
        self.assertEqual(self.sense.illustration_items[idx].href, href)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.illustration_items[idx]

    def test_sense_add_note(self):
        idx = len(self.sense.note_items) if self.sense.note_items else 0
        self.sense.add_note()
        len_after = len(self.sense.note_items)
        self.assertEqual(len_after - 1, idx)
        self.assertIsInstance(self.sense.note_items[idx], lexicon.Note)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.note_items[idx]

    def test_sense_add_relation(self):
        idx = len(self.sense.relation_items) if self.sense.relation_items else 0
        self.sense.add_relation()
        self.assertEqual(len(self.sense.relation_items), 1)
        self.assertIsInstance(self.sense.relation_items[idx], lexicon.Relation)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.relation_items[idx]

    def test_sense_add_reversal(self):
        idx = len(self.sense.reversal_items) if self.sense.reversal_items else 0
        self.sense.add_reversal()
        len_after = len(self.sense.reversal_items)
        self.assertEqual(len_after - 1, idx)
        self.assertIsInstance(self.sense.reversal_items[idx], lexicon.Reversal)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.reversal_items[idx]

    def test_sense_add_subsense(self):
        idx = len(self.sense.subsense_items) if self.sense.subsense_items else 0
        self.sense.add_subsense()
        self.assertEqual(len(self.sense.subsense_items), 1)
        self.assertIsInstance(self.sense.subsense_items[idx], lexicon.Sense)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)
        del self.sense.subsense_items[idx]

    def test_sense_set_definition(self):
        self.sense.set_definition({"en": "english text", "sg": "atëne ti sängö"})
        self.assertEqual(len(self.sense.definition.form_items), 2)
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)

    def test_sense_set_grammatical_info(self):
        self.sense.set_grammatical_info("Nom")
        self.assertEqual(str(self.sense.grammatical_info), "Nom")
        self.assertIsNotNone(self.sense.date_modified)
        self.assertNotEqual(self.sense.date_modified, self.sense_last_modified)



# flake8: noqa: E501

import unittest

from lift_utils import base, config, datatypes, utils

LIFT_VERSION = "0.13"


class TestAnnotation(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        name = "annotation-name"
        value = "annotation-value"
        who = "Annotation Who"
        when = "0000-00-00"
        obj = base.Annotation(name=name, value=value, who=who, when=when)

        xml = f"""
        <annotation name="{name}" value="{value}" who="{who}" when="{when}"/>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestField(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = base.Field(field_type="field-type")
        obj.date_created = datatypes.DateTime("0000-00-00")
        obj.date_modified = datatypes.DateTime("0000-00-00")
        obj.annotation_items = [
            base.Annotation(
                name="",
                value="",
            ),
            base.Annotation(
                name="",
                value="",
            ),
        ]
        obj.trait_items = [
            base.Trait(name="", value=""),
            base.Trait(name="", value=""),
        ]
        obj.form_items = [
            base.Form(lang="", text=""),
            base.Form(lang="", text=""),
        ]

        xml = f"""
        <field dateCreated="{obj.date_created}" dateModified="{obj.date_modified}" type="{obj.type}">
            <annotation name="" value=""/>
            <annotation name="" value=""/>
            <form lang="">
                <text/>
            </form>
            <form lang="">
                <text/>
            </form>
            <trait name="" value=""/>
            <trait name="" value=""/>
        </field>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestForm(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = base.Form(lang="en", text="")
        obj.annotation_items = [
            base.Annotation(
                name="",
                value="",
            ),
            base.Annotation(
                name="",
                value="",
            ),
        ]

        xml = """
        <form lang="en">
            <annotation name="" value=""/>
            <annotation name="" value=""/>
            <text/>
        </form>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestGloss(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = base.Gloss(lang="", text="")
        obj.trait_items = [
            base.Trait(name="", value=""),
            base.Trait(name="", value=""),
        ]

        xml = """
        <gloss lang="">
            <text/>
            <trait name="" value=""/>
            <trait name="" value=""/>
        </gloss>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestSpan(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        text = "test text"
        obj = base.Span(text=text)
        obj.lang = datatypes.Lang("en")
        obj.href = datatypes.URL("https://example.com")
        obj.class_ = "TestClass"
        obj.tail = datatypes.PCData(" a tail.")

        xml = f"""
        <xml>
        <span class="{obj.class_}" href="{obj.href}" lang="{obj.lang}" >{text}</span>{obj.tail}</xml>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_control = xml_tree_control.getchildren()[0]
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestText(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = base.Text("Test text")
        obj.span_items = [
            base.Span(text=""),
            base.Span(text=""),
        ]

        xml = """
        <text>Test text<span/><span/></text>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestTrait(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = base.Trait(name="TraitName", value="TraitValue")
        obj.id = datatypes.Key("TraitID")
        obj.annotation_items = [
            base.Annotation(name="", value=""),
            base.Annotation(name="", value=""),
        ]

        xml = f"""
        <trait name="TraitName" value="TraitValue" id="{obj.id}">
            <annotation name="" value=""/>
            <annotation name="" value=""/>
        </trait>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None


class TestURLRef(unittest.TestCase):
    def setUp(self):
        config.LIFT_VERSION = LIFT_VERSION

    def test_xml(self):
        obj = base.URLRef(href="https://example.com")
        obj.label = base.Multitext()

        xml = """
        <urlref href="https://example.com">
            <label/>
        </urlref>
        """
        xml_tree_control = utils.xmlstring_to_etree(xml)
        xml_tree_gen = obj._to_xml_tree()
        for tree in (xml_tree_control, xml_tree_gen):
            utils.sort_xml(tree)
        self.assertEqual(
            utils.etree_to_xmlstring(xml_tree_gen),
            utils.etree_to_xmlstring(xml_tree_control),
        )

    def tearDown(self):
        config.LIFT_VERSION = None
